import express from 'express'
import jwt from 'jsonwebtoken'
import cors from 'cors'
import { applyMiddleware } from 'graphql-middleware'
import { buildSubgraphSchema } from '@apollo/federation'
import { gql } from 'apollo-server'
import { createServer } from 'http'
import { ApolloServer } from 'apollo-server-express'
import { WebSocketServer } from 'ws'
import { ApolloServerPluginDrainHttpServer } from '@apollo/server/plugin/drainHttpServer'
import { useServer } from 'graphql-ws/lib/use/ws'
import { PubSub } from 'graphql-subscriptions'

const pubSub = new PubSub()

const initializeWebsocketServer = async (
  models,
  typeDefs,
  resolvers,
  middlewares,
  port
) => {
  console.log(port, '============================')
  const schemaWithoutMiddleware = buildSubgraphSchema([
    {
      typeDefs: gql`
        ${typeDefs}
      `,
      resolvers: { ...resolvers },
      resolverValidationOptions: { requireResolversForResolveType: false },
    },
  ])

  const schema = applyMiddleware(schemaWithoutMiddleware, ...middlewares)

  const validateToken = (authToken) => {
    const token = authToken.slice(7, authToken.length)
    return new Promise((resolve, reject) => {
      jwt.verify(token, process.env.JWT_SECRET, {}, (err, decoded) => {
        if (err) {
          console.log(err)
          reject(err)
        }
        return resolve(decoded)
      })
    })
  }
  const app = express()
  const httpServer = createServer(app)

  const wsServer = new WebSocketServer({
    server: httpServer,
    path: '/graphql',
  })

  const getDynamicContext = async (ctx) => {
    if (ctx.connectionParams.Authorization) {
      const userData = await validateToken(ctx.connectionParams.Authorization)
      console.log(userData, '========userData')
      return userData
    }
  }

  const serverCleanup = useServer(
    {
      schema,
      context: async (ctx, msg, args) => {
        return getDynamicContext(ctx, msg, args)
      },
    },
    wsServer
  )
  const server = new ApolloServer({
    schema,
    plugins: [
      ApolloServerPluginDrainHttpServer({ httpServer }),
      {
        async serverWillStart() {
          return {
            async drainServer() {
              await serverCleanup.dispose()
            },
          }
        },
      },
    ],
    context: async ({ req }) => {
      return { models, pubSub, req }
    },
  })

  await server.start()
  app.use('/graphql', cors())

  server.applyMiddleware({ app })
  httpServer.listen(port, () => {
    console.log(
      `🚀 Subscription endpoint ready at http://localhost:${port}/graphql`
    )
  })
}

export { initializeWebsocketServer, pubSub }
